import numpy as np
from moviepy.video.VideoClip import VideoClip, ColorClip
from moviepy.audio.AudioClip import CompositeAudioClip

#  CompositeVideoClip

class CompositeVideoClip(VideoClip):

    """ 
    
    A VideoClip made of other videoclips displayed together. This is the
    base class for most compositions.
    
    Parameters

    size
      The size (height x width) of the final clip.

    clips
      A list of videoclips. Each clip of the list will
      be displayed below the clips appearing after it in the list.
      For each clip:
       
      - The attribute ``pos`` determines where the clip is placed.
          See ``VideoClip.set_pos``
      - The mask of the clip determines which parts are visible.
        
      Finally, if all the clips in the list have their ``duration``
      attribute set, then the duration of the composite video clip
      is computed automatically

    bg_color
      Color for the unmasked and unfilled regions. Set to None for these
      regions to be transparent (will be slower).
    
    If all clips with a fps attribute have the same fps, it becomes the fps of
    the result.

    """

    def __init__(self, clips, size=None, bg_color=None, use_bgclip=False,
                 ismask=False):

        if size is None:
            size = clips[0].size
        
        if bg_color is None:
            bg_color = 0.0 if ismask else (0, 0, 0)

        if use_bgclip and (clips[0].mask is None):
            transparent = False
        else:
            transparent = (bg_color is None)
        
        fps_list = list(set([c.fps for c in clips if hasattr(c,'fps')]))
        if len(fps_list)==1:
            self.fps= fps_list[0]

        VideoClip.__init__(self)
        
        self.size = size
        self.ismask = ismask
        self.clips = clips
        self.bg_color = bg_color

        if use_bgclip:
            self.bg = clips[0]
            self.clips = clips[1:]
        else:
            self.clips = clips
            self.bg = ColorClip(size, col=self.bg_color)

        
        
        # compute duration
        ends = [c.end for c in self.clips]
        if not any([(e is None) for e in ends]):
            self.duration = max(ends)
            self.end = max(ends)

        # compute audio
        audioclips = [v.audio for v in self.clips if v.audio is not None]
        if len(audioclips) > 0:
            self.audio = CompositeAudioClip(audioclips)

        # compute mask if necessary
        if transparent:
            maskclips = [c.mask.set_pos(c.pos) for c in self.clips
                         if c.mask is not None]
            if maskclips != []:
                self.mask = CompositeVideoClip(maskclips,self.size, ismask=True,
                                               bg_color=0)

        def make_frame(t):
            """ The clips playing at time `t` are blitted over one
                another. """

            f = self.bg.get_frame(t)
            for c in self.playing_clips(t):
                    f = c.blit_on(f, t)
            return f

        self.make_frame = make_frame

    def playing_clips(self, t=0):
        """ Returns a list of the clips in the composite clips that are
            actually playing at the given time `t`. """
        return [c for c in self.clips if c.is_playing(t)]



def clips_array(array, rows_widths=None, cols_widths=None,
                bg_color = None):

    """

    rows_widths
      widths of the different rows in pixels. If None, is set automatically.

    cols_widths
      widths of the different colums in pixels. If None, is set automatically.

    cols_widths
    
    bg_color
       Fill color for the masked and unfilled regions. Set to None for these
       regions to be transparent (will be slower).

    """
    
    array = np.array(array)
    sizes_array = np.array([[c.size for c in line] for line in array])
    
    # find row width and col_widths automatically if not provided
    if rows_widths is None:
        rows_widths = sizes_array[:,:,1].max(axis=1)
    if cols_widths is None:
        cols_widths = sizes_array[:,:,0].max(axis=0)
    
    xx = np.cumsum([0]+list(cols_widths)) 
    yy = np.cumsum([0]+list(rows_widths))
    
    for j,(x,cw) in list(enumerate(zip(xx[:-1],cols_widths))):
        for i,(y,rw) in list(enumerate(zip(yy[:-1],rows_widths))):
            clip = array[i,j]
            w,h = clip.size
            if (w < cw) or (h < rw):
                clip = (CompositeVideoClip([clip.set_pos('center')],
                                          size = (cw,rw),
                                          bg_color = bg_color).
                                     set_duration(clip.duration))
                
            array[i,j] = clip.set_pos((x,y))
                 
    return CompositeVideoClip(array.flatten(), size = (xx[-1],yy[-1]),
                              bg_color = bg_color)
    
    



""" This module contains everything that can help automatize
the cuts in MoviePy """

from moviepy.decorators import use_clip_fps_by_default
import numpy as np

@use_clip_fps_by_default
def find_video_period(clip,fps=None,tmin=.3):
    """ Finds the period of a video based on frames correlation """
    

    frame = lambda t: clip.get_frame(t).flatten()
    tt = np.arange(tmin,clip.duration,1.0/ fps)[1:]
    ref = frame(0)
    corrs = [ np.corrcoef(ref, frame(t))[0,1] for t in tt]
    return tt[np.argmax(corrs)]


class FrameMatch:

    def __init__(self, t1, t2, distance):
        self.t1, self.t2, self.distance = t1, t2, distance
        self.time_span = t2-t1

    def __str__(self):
        return '(%.04f, %.04f, %.04f)'%(self.t1, self.t2, self.time_span)

    def __repr__(self):
        return '(%.04f, %.04f, %.04f)'%(self.t1, self.t2, self.time_span)

    def __iter__(self):
        return [self.t1, self.t2, self.distance].__iter__()


class FrameMatches(list):

    def __init__(self, lst):

        list.__init__(self, sorted(lst, key=lambda e: e.distance))

    def best(self, n=None, percent=None):
        if percent is not None:
            n = len(self)*percent/100
        return FrameMatches(self[:n])
    
    def filter(self, fun):
        return FrameMatches(filter(fun, self))

    def save(self, filename):
        np.savetxt(np.array(self), matching_frames, fmt='%.03f', delimiter='\t')

    @staticmethod
    def load(filename):
        arr = np.loadtxt(filename)
        mfs = [FrameMatch(*e) for e in arr]

        



def find_matching_frames(clip, dist_thr, max_d, fps=None):
    """ Returns a list [(t1, t2, distance), ...] for all pairs of frames with
    (t2-t1 < max_d). Optimized routine. Quite fast.

    clip : a MoviePy video clip, possibly transformed/resized
    dist_thr: distance above which a match is rejected
    max_d: maximal duration (in seconds) between two matching frames
    fps: frames per second (default will be clip.fps)
    """ 
    
    N_pixels = clip.w * clip.h * 3
    dot_product = lambda F1, F2: (F1*F2).sum()/N_pixels
    F = {} # will store the frames and their mutual distances
    
    def distance(t1, t2):
        uv = dot_product(F[t1]['frame'], F[t2]['frame'])
        u, v = F[t1]['|F|sq'], F[t2]['|F|sq']
        return np.sqrt(u+v - 2*uv)
    
    matching_frames = [] # the final result.
    
    for (t,frame) in clip.iter_frames(with_times=True, progress_bar=True):
        
        flat_frame = 1.0*frame.flatten()
        F_norm_sq = dot_product(flat_frame, flat_frame)
        F_norm = np.sqrt(F_norm_sq)
        
        for t2 in F.keys():
            # forget old frames, add 't' to the others frames
            # check for early rejections based on differing norms
            if (t-t2) > max_d:
                F.pop(t2)
            else:
                F[t2][t] = {'min':abs(F[t2]['|F|'] - F_norm),
                            'max':F[t2]['|F|'] + F_norm}
                F[t2][t]['rejected']= (F[t2][t]['min'] > dist_thr)
        
        t_F = sorted(F.keys())
        
        F[t] = {'frame': flat_frame, '|F|sq': F_norm_sq, '|F|': F_norm}
                
        for i,t2 in enumerate(t_F):
            # Compare F(t) to all the previous frames
            
            if F[t2][t]['rejected']:
                continue
            
            dist = distance(t, t2)
            F[t2][t]['min'] = F[t2][t]['max'] = dist
            F[t2][t]['rejected']  = (dist >= dist_thr)
            
            for t3 in t_F[i+1:]:
                # For all the next times t3, use d(F(t), F(t2)) to
                # update the bounds on d(F(t), F(t3)). See if you can
                # conclude on wether F(t) and F(t3) match.
                t3t, t2t3 = F[t3][t], F[t2][t3]
                t3t['max'] = min(t3t['max'], dist+ t2t3['max'])
                t3t['min'] = max(t3t['min'], dist - t2t3['max'],
                                 t2t3['min'] - dist)
                                      
                if t3t['min'] > dist_thr:
                    t3t['rejected'] = True
    
        # Store all the good matches (t2,t)
        matching_frames += [(t1, t, F[t1][t]['min']) for t1 in F
                            if (t1!=t) and not F[t1][t]['rejected']]
                   
    return FrameMatches([FrameMatch(*e) for e in matching_frames])



@use_clip_fps_by_default
def detect_scenes(clip=None, luminosities=None, thr=10,
                  progress_bar=False, fps=None):
    
    """ Detects scenes of a clip based on luminosity changes.
    
    Note that for large clip this may take some time
    
    Returns
    --------
    cuts, luminosities
      cuts is a series of cuts [(0,t1), (t1,t2),...(...,tf)]
      luminosities are the luminosities computed for each
      frame of the clip.
    
    Parameters
    -----------
    
    clip
      A video clip. Can be None if a list of luminosities is
      provided instead. If provided, the luminosity of each
      frame of the clip will be computed. If the clip has no
      'fps' attribute, you must provide it.
    
    luminosities
      A list of luminosities, e.g. returned by detect_scenes
      in a previous run.
    
    thr
      Determines a threshold above which the 'luminosity jumps'
      will be considered as scene changes. A scene change is defined
      as a change between 2 consecutive frames that is larger than
      (avg * thr) where avg is the average of the absolute changes
      between consecutive frames.
      
    progress_bar
      We all love progress bars ! Here is one for you, in option.
      
    fps
      Must be provided if you provide no clip or a clip without
      fps attribute.
    
    
    
    
    """
        
    if luminosities is None:
        luminosities = [f.sum() for f in clip.iter_frames(
                             fps=fps, dtype='uint32', progress_bar=1)]
    
    luminosities = np.array(luminosities, dtype=float)
    if clip is not None:
        end = clip.duration
    else:
        end = len(luminosities)*(1.0/fps) 
    lum_diffs = abs(np.diff(luminosities))
    avg = lum_diffs.mean()
    luminosity_jumps = 1+np.array(np.nonzero(lum_diffs> thr*avg))[0]
    tt = [0]+list((1.0/fps) *luminosity_jumps) + [end]
    #print tt
    cuts = [(t1,t2) for t1,t2 in zip(tt,tt[1:])]
    return cuts, luminosities
